@@ -13,7 +13,7 @@
 
 package IO::Socket::SSL;
 
-our $VERSION = '2.005_1';
+our $VERSION = '2.015_001';
 
 use IO::Socket;
 use Net::SSLeay 1.46;
@@ -49,7 +49,8 @@ use constant SSL_OCSP_TRY_STAPLE  => 0b10000;
 # capabilities of underlying Net::SSLeay/openssl
 my $can_client_sni;  # do we support SNI on the client side
 my $can_server_sni;  # do we support SNI on the server side
-my $can_npn;         # do we support NPN
+my $can_npn;         # do we support NPN (obsolete)
+my $can_alpn;        # do we support ALPN
 my $can_ecdh;        # do we support ECDH key exchange
 my $can_ocsp;        # do we support OCSP
 my $can_ocsp_staple; # do we support OCSP stapling
@@ -57,6 +58,7 @@ BEGIN {
     $can_client_sni = Net::SSLeay::OPENSSL_VERSION_NUMBER() >= 0x01000000;
     $can_server_sni = defined &Net::SSLeay::get_servername;
     $can_npn        = defined &Net::SSLeay::P_next_proto_negotiated;
+    $can_alpn       = defined &Net::SSLeay::CTX_set_alpn_protos;
     $can_ecdh       = defined &Net::SSLeay::CTX_set_tmp_ecdh &&
 	# There is a regression with elliptic curves on 1.0.1d with 64bit
 	# http://rt.openssl.org/Ticket/Display.html?id=2975
@@ -89,9 +91,10 @@ my %DEFAULT_SSL_ARGS = (
     SSL_verifycn_publicsuffix => undef,  # fallback default list verification
     #SSL_verifycn_name => undef,   # use from PeerAddr/PeerHost - do not override in set_args_filter_hack 'use_defaults'
     SSL_npn_protocols => undef,    # meaning depends whether on server or client side
+    SSL_alpn_protocols => undef,   # list of protocols we'll accept/send, for example ['http/1.1','spdy/3.1']
     SSL_cipher_list =>
 	'EECDH+AESGCM+ECDSA EECDH+AESGCM EECDH+ECDSA +AES256 EECDH EDH+AESGCM '.
-	'EDH ALL +SHA +3DES +RC4 !LOW !EXP !eNULL !aNULL !DES !MD5 !PSK !SRP',
+	'EDH ALL +SHA +3DES !RC4 !LOW !EXP !eNULL !aNULL !DES !MD5 !PSK !SRP',
 );
 
 my %DEFAULT_SSL_CLIENT_ARGS = (
@@ -253,6 +256,7 @@ use vars qw(@ISA $SSL_ERROR @EXPORT);
 my @caller_force_inet4; # in case inet4 gets forced we store here who forced it
 
 my $IOCLASS;
+my $family_key; # 'Domain'||'Family'
 BEGIN {
     # declare @ISA depending of the installed socket class
 
@@ -272,15 +276,17 @@ BEGIN {
     };
 
     # try IO::Socket::IP or IO::Socket::INET6 for IPv6 support
+    $family_key = 'Domain'; # traditional
     if ( $ip6 ) {
-	# if we have IO::Socket::IP >= 0.20 we will use this in preference
+	# if we have IO::Socket::IP >= 0.31 we will use this in preference
 	# because it can handle both IPv4 and IPv6
 	if ( eval { 
 	    require IO::Socket::IP; 
-	    IO::Socket::IP->VERSION(0.20) && IO::Socket::IP->VERSION != 0.30; 
+	    IO::Socket::IP->VERSION(0.31)
 	}) {
 	    @ISA = qw(IO::Socket::IP);
 	    constant->import( CAN_IPV6 => "IO::Socket::IP" );
+	    $family_key = 'Family';
 	    $IOCLASS = "IO::Socket::IP";
 
 	# if we have IO::Socket::INET6 we will use this not IO::Socket::INET
@@ -312,7 +318,12 @@ BEGIN {
 
 sub DEBUG {
     $DEBUG or return;
-    my (undef,$file,$line) = caller;
+    my (undef,$file,$line,$sub) = caller(1);
+    if ($sub =~m{^IO::Socket::SSL::(?:error|(_internal_error))$}) {
+	(undef,$file,$line) = caller(2) if $1;
+    } else {
+	(undef,$file,$line) = caller;
+    }
     my $msg = shift;
     $file = '...'.substr( $file,-17 ) if length($file)>20;
     $msg = sprintf $msg,@_ if @_;
@@ -416,7 +427,7 @@ sub import {
 		    @ISA = ( CAN_IPV6 );
 		    warn "IPv6 support re-enabled in __PACKAGE__, got disabled in file $caller_force_inet4[1] line $caller_force_inet4[2]";
 		} else {
-		    die "INET6 is not supported, install IO::Socket::INET6";
+		    die "INET6 is not supported, install IO::Socket::IP";
 		}
 	    }
 	} elsif ( /^:?debug(\d+)/ ) {
@@ -434,6 +445,13 @@ my %SSL_OBJECT;
 my %CREATED_IN_THIS_THREAD;
 sub CLONE { %CREATED_IN_THIS_THREAD = (); }
 
+
+# we have callbacks associated with contexts, but have no way to access the
+# current SSL object from these callbacks. To work around this
+# CURRENT_SSL_OBJECT will be set before calling Net::SSLeay::{connect,accept}
+# and reset afterwards, so we have access to it inside _internal_error.
+my $CURRENT_SSL_OBJECT;
+
 # You might be expecting to find a new() subroutine here, but that is
 # not how IO::Socket::INET works.  All configuration gets performed in
 # the calls to configure() and either connect() or accept().
@@ -458,8 +476,9 @@ sub configure {
 
     $self->configure_SSL($arg_hash) || return;
 
-    $self->SUPER::configure($arg_hash)
-	|| return $self->_internal_error("@ISA configuration failed");
+    $arg_hash->{$family_key} ||= $arg_hash->{Domain} || $arg_hash->{Family};
+    return $self->_internal_error("@ISA configuration failed",0)
+	if ! $self->SUPER::configure($arg_hash);
 
     $self->blocking(0) if defined $blocking && !$blocking;
     return $self;
@@ -520,7 +539,7 @@ sub _skip_rw_error {
 	return $err;
     }
     $! ||= EWOULDBLOCK;
-    ${*$self}{'_SSL_last_err'} = $SSL_ERROR if ref($self);
+    ${*$self}{_SSL_last_err} = [$SSL_ERROR,4] if ref($self);
     Net::SSLeay::ERR_clear_error();
     return 0;
 }
@@ -565,7 +584,7 @@ sub connect_SSL {
 	my $arg_hash = ${*$self}{'_SSL_arguments'};
 
 	my $fileno = ${*$self}{'_SSL_fileno'} = fileno($self);
-	return $self->_internal_error("Socket has no fileno")
+	return $self->_internal_error("Socket has no fileno",9)
 	    if ! defined $fileno;
 
 	$ctx = ${*$self}{'_SSL_ctx'};  # Reference to real context
@@ -600,7 +619,7 @@ sub connect_SSL {
 	    }
 	} elsif ( $arg_hash->{SSL_hostname} ) {
 	    return $self->_internal_error(
-		"Client side SNI not supported for this openssl");
+		"Client side SNI not supported for this openssl",9);
 	} else {
 	    $DEBUG>=2 && DEBUG("not using SNI because openssl is too old");
 	}
@@ -657,10 +676,13 @@ sub connect_SSL {
     }
 
     my $start = defined($timeout) && time();
-    for my $dummy (1) {
-	#DEBUG( 'calling ssleay::connect' );
+    {
+	$SSL_ERROR = undef;
+	$CURRENT_SSL_OBJECT = $self;
+	$DEBUG>=3 && DEBUG("call Net::SSLeay::connect" );
 	my $rv = Net::SSLeay::connect($ssl);
-	$DEBUG>=3 && DEBUG("Net::SSLeay::connect -> $rv" );
+	$CURRENT_SSL_OBJECT = undef;
+	$DEBUG>=3 && DEBUG("done Net::SSLeay::connect -> $rv" );
 	if ( $rv < 0 ) {
 	    if ( my $err = $self->_skip_rw_error( $ssl,$rv )) {
 		$self->error("SSL connect attempt failed");
@@ -719,7 +741,10 @@ sub connect_SSL {
     # ssl connect successful
     delete ${*$self}{'_SSL_opening'};
     ${*$self}{'_SSL_opened'}=1;
-    $self->blocking(1) if defined($timeout); # was blocking before
+    if (defined($timeout)) {
+	$self->blocking(1); # reset back to blocking
+	$! = undef; # reset errors from non-blocking
+    }
 
     $ctx ||= ${*$self}{'_SSL_ctx'};
 
@@ -738,11 +763,11 @@ sub connect_SSL {
 	    # definitly revoked
 	    $DEBUG>=3 && DEBUG("got OCSP revocation with stapling: %s",
 		$ocsp_result->[1]);
-	    $self->_internal_error($ocsp_result->[1]);
+	    $self->_internal_error($ocsp_result->[1],5);
 	    return $self->fatal_ssl_error();
 	}
     } elsif ( $ctx->{ocsp_mode} & SSL_OCSP_MUST_STAPLE ) {
-	$self->_internal_error("did not receive the required stapled OCSP response");
+	$self->_internal_error("did not receive the required stapled OCSP response",5);
 	return $self->fatal_ssl_error();
     }
 
@@ -810,7 +835,7 @@ sub accept {
 	    };
 	    $DEBUG>=2 && DEBUG('will not start SSL handshake yet');
 	    return wantarray ? ($socket, getpeername($socket) ) : $socket
-	};
+	}
     }
 
     $self->accept_SSL($socket) || return;
@@ -837,7 +862,7 @@ sub accept_SSL {
 	}
 
 	my $fileno = ${*$socket}{'_SSL_fileno'} = fileno($socket);
-	return $socket->_internal_error("Socket has no fileno")
+	return $socket->_internal_error("Socket has no fileno",9)
 	    if ! defined $fileno;
 
 	$ssl = ${*$socket}{_SSL_object} =
@@ -868,8 +893,11 @@ sub accept_SSL {
     }
 
     my $start = defined($timeout) && time();
-    for my $dummy (1) {
+    {
+	$SSL_ERROR = undef;
+	$CURRENT_SSL_OBJECT = $self;
 	my $rv = Net::SSLeay::accept($ssl);
+	$CURRENT_SSL_OBJECT = undef;
 	$DEBUG>=3 && DEBUG( "Net::SSLeay::accept -> $rv" );
 	if ( $rv < 0 ) {
 	    if ( my $err = $socket->_skip_rw_error( $ssl,$rv )) {
@@ -889,7 +917,7 @@ sub accept_SSL {
 		my $vec = '';
 		vec($vec,$socket->fileno,1) = 1;
 		$rv =
-		    $SSL_ERROR == SSL_WANT_READ ? select( $vec,undef,undef,$timeout) :
+		    $SSL_ERROR == SSL_WANT_READ  ? select( $vec,undef,undef,$timeout) :
 		    $SSL_ERROR == SSL_WANT_WRITE ? select( undef,$vec,undef,$timeout) :
 		    undef;
 	    } else {
@@ -911,7 +939,7 @@ sub accept_SSL {
 	    redo;
 
 	} elsif ( $rv == 0 ) {
-	    $socket->error("SSL connect accept failed because of handshake problems" );
+	    $socket->error("SSL accept attempt failed because of handshake problems" );
 	    delete ${*$self}{'_SSL_opening'};
 	    ${*$socket}{'_SSL_opened'} = -1;
 	    return $socket->fatal_ssl_error();
@@ -922,7 +950,10 @@ sub accept_SSL {
     # socket opened
     delete ${*$self}{'_SSL_opening'};
     ${*$socket}{'_SSL_opened'} = 1;
-    $socket->blocking(1) if defined($timeout); # was blocking before
+    if (defined($timeout)) {
+	$socket->blocking(1); # reset back to blocking
+	$! = undef; # reset errors from non-blocking
+    }
 
     tie *{$socket}, "IO::Socket::SSL::SSL_HANDLE", $socket;
 
@@ -1001,7 +1032,7 @@ sub _generic_write {
     my $buf_len = length($$buffer);
     $length ||= $buf_len;
     $offset ||= 0;
-    return $self->_internal_error("Invalid offset for SSL write")
+    return $self->_internal_error("Invalid offset for SSL write",9)
 	if $offset>$buf_len;
     return 0 if ($offset == $buf_len);
 
@@ -1066,6 +1097,7 @@ sub getc {
 
 sub readline {
     my $self = shift;
+    ${*$self}{_SSL_object} or return $self->SUPER::getline;
 
     if ( not defined $/ or wantarray) {
 	# read all and split
@@ -1137,7 +1169,7 @@ sub readline {
 	    $buf .= $pb
 	} else {
 	    return $buf eq '' ? ():$buf;
-	};
+	}
 	if ( !$eod ) {
 	    my $pos = index( $buf,$delim0 );
 	    if ( $pos<0 ) {
@@ -1303,7 +1335,7 @@ sub fileno {
 sub _get_ssl_object {
     my $self = shift;
     return ${*$self}{'_SSL_object'} ||
-	IO::Socket::SSL->_internal_error("Undefined SSL object");
+	IO::Socket::SSL->_internal_error("Undefined SSL object",9);
 }
 
 # _get_ctx_object is for internal use ONLY!
@@ -1315,7 +1347,7 @@ sub _get_ctx_object {
 
 # default error for undefined arguments
 sub _invalid_object {
-    return IO::Socket::SSL->_internal_error("Undefined IO::Socket::SSL object");
+    return IO::Socket::SSL->_internal_error("Undefined IO::Socket::SSL object",9);
 }
 
 
@@ -1326,18 +1358,18 @@ sub pending {
 
 sub start_SSL {
     my ($class,$socket) = (shift,shift);
-    return $class->_internal_error("Not a socket") if ! ref($socket);
+    return $class->_internal_error("Not a socket",9) if ! ref($socket);
     my $arg_hash = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
     my %to = exists $arg_hash->{Timeout} ? ( Timeout => delete $arg_hash->{Timeout} ) :();
     my $original_class = ref($socket);
     if ( ! $original_class ) {
 	$socket = ($original_class = $ISA[0])->new_from_fd($socket,'<+')
 	    or return $class->_internal_error(
-	    "creating $original_class from file handle failed");
+	    "creating $original_class from file handle failed",9);
     }
     my $original_fileno = (UNIVERSAL::can($socket, "fileno"))
 	? $socket->fileno : CORE::fileno($socket);
-    return $class->_internal_error("Socket has no fileno")
+    return $class->_internal_error("Socket has no fileno",9)
 	if ! defined $original_fileno;
 
     bless $socket, $class;
@@ -1556,7 +1588,7 @@ if ( defined &Net::SSLeay::get_peer_cert_chain
 	my $publicsuffix = shift;
 	if ( ! ref($scheme) ) {
 	    $DEBUG>=3 && DEBUG( "scheme=$scheme cert=$cert" );
-	    $scheme = $scheme{$scheme} or croak "scheme $scheme not defined";
+	    $scheme = $scheme{$scheme} || croak("scheme $scheme not defined");
 	}
 
 	return 1 if ! %$scheme; # 'none'
@@ -1577,6 +1609,7 @@ if ( defined &Net::SSLeay::get_peer_cert_chain
 	my $ipn;
 	if ( CAN_IPV6 and $identity =~m{:} ) {
 	    # no IPv4 or hostname have ':'  in it, try IPv6.
+	    $identity =~m{[^\da-fA-F:\.]} and return; # invalid characters in name
 	    $ipn = inet_pton(AF_INET6,$identity) or return; # invalid name
 	} elsif ( my @ip = $identity =~m{^(\d+)(?:\.(\d+)\.(\d+)\.(\d+)|[\d\.]*)$} ) {
 	    # check for invalid IP/hostname
@@ -1735,7 +1768,8 @@ if ($can_ocsp) {
 
 sub errstr {
     my $self = shift;
-    return (ref($self) ? ${*$self}{'_SSL_last_err'} : $SSL_ERROR) || '';
+    my $oe = ref($self) && ${*$self}{_SSL_last_err};
+    return $oe ? $oe->[0] : $SSL_ERROR || '';
 }
 
 sub fatal_ssl_error {
@@ -1761,35 +1795,55 @@ sub get_ssleay_error {
     return Net::SSLeay::print_errs('SSL error: ') || '';
 }
 
-# internal errors, e.g unsupported features etc
+# internal errors, e.g unsupported features, hostname check failed etc
+# _SSL_last_err contains severity so that on error chains we can decide if one
+# error should replace the previous one or if this is just a less specific
+# follow-up error, e.g. configuration failed because certificate failed because
+# hostname check went wrong:
+# 0 - fallback errors
+# 4 - errors bubbled up from OpenSSL (sub error, r/w error)
+# 5 - hostname or OCSP verification failed
+# 9 - fatal problems, e.g. missing feature, no fileno...
+# _SSL_last_err and SSL_ERROR are only replaced if the error has a higher
+# severity than the previous one
+
 sub _internal_error {
-    my ($self, $error, $destroy_socket) = @_;
-    $SSL_ERROR = dualvar( -1, $error );
-    $DEBUG && DEBUG($error);
-    ${*$self}{'_SSL_last_err'} = $SSL_ERROR if (ref($self));
+    my ($self, $error, $severity) = @_;
+    $error = dualvar( -1, $error );
+    $self = $CURRENT_SSL_OBJECT if !ref($self) && $CURRENT_SSL_OBJECT;
+    if (ref($self)) {
+	my $oe = ${*$self}{_SSL_last_err};
+	if (!$oe || $oe->[1] <= $severity) {
+	    ${*$self}{_SSL_last_err} = [$error,$severity];
+	    $SSL_ERROR = $error;
+	    $DEBUG && DEBUG("local error: $error");
+	} else {
+	    $DEBUG && DEBUG("ignoring less severe local error '$error', keep '$oe->[0]'");
+	}
+    } else {
+	$SSL_ERROR = $error;
+	$DEBUG && DEBUG("global error: $error");
+    }
     return;
 }
 
 # OpenSSL errors
 sub error {
-    my ($self, $error, $destroy_socket) = @_;
+    my ($self, $error) = @_;
     my @err;
     while ( my $err = Net::SSLeay::ERR_get_error()) {
 	push @err, Net::SSLeay::ERR_error_string($err);
 	$DEBUG>=2 && DEBUG( $error."\n".$self->get_ssleay_error());
     }
     $error .= ' '.join(' ',@err) if @err;
-    if ($error) {
-	$SSL_ERROR = dualvar( -1, $error );
-	$DEBUG && DEBUG($error);
-	${*$self}{'_SSL_last_err'} = $SSL_ERROR if (ref($self));
-    }
+    return $self->_internal_error($error,4) if $error;
     return;
 }
 
 sub can_client_sni { return $can_client_sni }
 sub can_server_sni { return $can_server_sni }
 sub can_npn        { return $can_npn }
+sub can_alpn       { return $can_alpn }
 sub can_ecdh       { return $can_ecdh }
 sub can_ipv6       { return CAN_IPV6 }
 sub can_ocsp       { return $can_ocsp }
@@ -1884,11 +1938,18 @@ sub set_args_filter_hack {
 
 sub next_proto_negotiated {
     my $self = shift;
-    return $self->_internal_error("NPN not supported in Net::SSLeay") if ! $can_npn;
+    return $self->_internal_error("NPN not supported in Net::SSLeay",9) if ! $can_npn;
     my $ssl = $self->_get_ssl_object || return;
     return Net::SSLeay::P_next_proto_negotiated($ssl);
 }
 
+sub alpn_selected {
+    my $self = shift;
+    return $self->_internal_error("ALPN not supported in Net::SSLeay",9) if ! $can_alpn;
+    my $ssl = $self->_get_ssl_object || return;
+    return Net::SSLeay::P_alpn_selected($ssl);
+}
+
 sub opened {
     my $self = shift;
     return IO::Handle::opened($self) && ${*$self}{'_SSL_opened'};
@@ -2060,7 +2121,7 @@ sub new {
 	    if ( ! $host ) {
 		if ( $vcn_scheme ) {
 		    IO::Socket::SSL->_internal_error(
-			"Cannot determine peer hostname for verification" );
+			"Cannot determine peer hostname for verification",8);
 		    return 0;
 		}
 		warn "Cannot determine hostname of peer for verification. ".
@@ -2096,7 +2157,7 @@ WARN
 	    }
 	    if ( ! $rv ) {
 		IO::Socket::SSL->_internal_error(
-		    "hostname verification failed" );
+		    "hostname verification failed",5);
 	    }
 	    return $rv;
 	};
@@ -2130,7 +2191,7 @@ WARN
 	$ver eq 'TLSv1_1' ? 'CTX_tlsv1_1_new' :
 	$ver eq 'TLSv1_2' ? 'CTX_tlsv1_2_new' :
 	'CTX_new'
-    ) or return IO::Socket::SSL->_internal_error("SSL Version $ver not supported");
+    ) or return IO::Socket::SSL->_internal_error("SSL Version $ver not supported",9);
 
     # For SNI in server mode we need a separate context for each certificate.
     my %ctx;
@@ -2185,7 +2246,7 @@ WARN
 	    SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER|SSL_MODE_ENABLE_PARTIAL_WRITE);
 
 	if ( my $proto_list = $arg_hash->{SSL_npn_protocols} ) {
-	    return IO::Socket::SSL->_internal_error("NPN not supported in Net::SSLeay")
+	    return IO::Socket::SSL->_internal_error("NPN not supported in Net::SSLeay",9)
 		if ! $can_npn;
 	    if($arg_hash->{SSL_server}) {
 		# on server side SSL_npn_protocols means a list of advertised protocols
@@ -2197,6 +2258,16 @@ WARN
 	    }
 	}
 
+	if ( my $proto_list = $arg_hash->{SSL_alpn_protocols} ) {
+	    return IO::Socket::SSL->_internal_error("ALPN not supported in Net::SSLeay",9)
+		if ! $can_alpn;
+	    if($arg_hash->{SSL_server}) {
+		Net::SSLeay::CTX_set_alpn_select_cb($ctx, $proto_list);
+	    } else {
+		Net::SSLeay::CTX_set_alpn_protos($ctx, $proto_list);
+	    }
+	}
+
 	# Try to apply SSL_ca even if SSL_verify_mode is 0, so that they can be
 	# used to verify OCSP responses.
 	# If applying fails complain only if verify_mode != VERIFY_NONE.
@@ -2231,6 +2302,24 @@ WARN
 	    }
 	}
 
+	if ($arg_hash->{SSL_server}
+	    && ($verify_mode & Net::SSLeay::VERIFY_PEER())) {
+	    if ($arg_hash->{SSL_client_ca}) {
+		for (@{$arg_hash->{SSL_client_ca}}) {
+		    return IO::Socket::SSL->error(
+			"Failed to add certificate to client CA list") if
+			! Net::SSLeay::CTX_add_client_CA($ctx,$_);
+		}
+	    }
+	    if ($arg_hash->{SSL_client_ca_file}) {
+		my $list = Net::SSLeay::load_client_CA_file(
+		    $arg_hash->{SSL_client_ca_file}) or
+		    return IO::Socket::SSL->error(
+		    "Failed to load certificate to client CA list");
+		Net::SSLeay::CTX_set_client_CA_list($ctx,$list);
+	    }
+	}
+
 	if ($arg_hash->{'SSL_check_crl'}) {
 	    Net::SSLeay::X509_STORE_set_flags(
 		Net::SSLeay::CTX_get_cert_store($ctx),
@@ -2293,8 +2382,8 @@ WARN
 		# don't free @chain, because CTX_add_extra_chain_cert
 		# did not duplicate the certificates
 	    }
-	    $havecert or return
-		IO::Socket::SSL->error("Failed to use certificate file");
+	    $havecert or return IO::Socket::SSL->error(
+		"Failed to load certificate from file (no PEM, DER or PKCS12)");
 	}
 
 	if (!$havecert || $havekey) {
@@ -2312,9 +2401,8 @@ WARN
 		    last;
 		}
 	    }
-	}
-	if ($havecert && !$havekey) {
-	    return IO::Socket::SSL->error("Failed to use private key");
+	    $havekey or return IO::Socket::SSL->error(
+		"Failed to load key from file (no PEM or DER)");
 	}
 
 	# replace arg_hash with created context
@@ -2346,7 +2434,7 @@ WARN
 
 	if ( my $curve = $arg_hash->{SSL_ecdh_curve} ) {
 	    return IO::Socket::SSL->_internal_error(
-		"ECDH curve needs Net::SSLeay>=1.56 and OpenSSL>=1.0")
+		"ECDH curve needs Net::SSLeay>=1.56 and OpenSSL>=1.0",9)
 		if ! $can_ecdh;
 	    if ( $curve !~ /^\d+$/ ) {
 		# name of curve, find NID
@@ -2371,7 +2459,7 @@ WARN
     if ( my $fp = $arg_hash->{SSL_fingerprint} ) {
 	for( ref($fp) ? @$fp : $fp) {
 	    my ($algo,$digest) = m{^([\w-]+)\$([a-f\d:]+)$}i;
-	    return IO::Socket::SSL->_internal_error("invalid fingerprint '$_'")
+	    return IO::Socket::SSL->_internal_error("invalid fingerprint '$_'",9)
 		if ! $algo;
 	    $algo = lc($algo);
 	    ( $digest = lc($digest) ) =~s{:}{}g;
@@ -2524,7 +2612,7 @@ WARN
     my $ctx = $ctx{''} || (values %ctx)[0];
     if (keys(%ctx) > 1 || ! exists $ctx{''}) {
 	$can_server_sni or return IO::Socket::SSL->_internal_error(
-	    "Server side SNI not supported for this openssl/Net::SSLeay");
+	    "Server side SNI not supported for this openssl/Net::SSLeay",9);
 
 	Net::SSLeay::CTX_set_tlsext_servername_callback($ctx, sub {
 	    my $ssl = shift;
